# -*- coding: utf-8 -*-
from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from h5sh import __version__
import sys


def run(inp, debug=False):
    from h5sh.state import State
    from h5sh.console import Console
    with State(inp) as state:
        console = Console(state)
        console.debug = debug
        console.interact()


def main(argv=None):
    from argparse import ArgumentParser

    # Load version
    this_version = __version__

    h5err = None
    try:
        from h5py import version as h5version
    except ImportError as e:
        h5err = e
        h5py_vers = hdf5_vers = "UNAVAILABLE"
    else:
        h5py_vers = h5version.version
        hdf5_vers = h5version.hdf5_version

    py_vers = "{v.major}.{v.minor}.{v.micro}".format(v=sys.version_info)
    version_str = "h5sh {} [Python {}] [h5py {}] [HDF5 {}]".format(
        this_version, py_vers, h5py_vers, hdf5_vers)

    # Create parser
    parser = ArgumentParser(
        prog="h5sh",
        description="shell-like interface to interacting with HDF5 files")
    parser.add_argument('inp',
                        help="Name of the HDF5 input file")
    parser.add_argument('--version', action="version",
                        version=version_str)
    parser.add_argument('-g', '--debug', action="store_true")

    args = parser.parse_args(argv)

    if h5err is not None:
        # h5py is unavailable
        parser.error("The 'h5py' Python package is required to run this "
                     "utility: " + str(h5err))
        sys.exit(2)

    # Print the version string at the top
    print(version_str)

    # Run the program
    run(**vars(args))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Functions for querying data on an HDF5 file."""

from __future__ import (division, absolute_import, print_function, )
#-----------------------------------------------------------------------------#
import numpy as np
from pprint import pformat
import sys
from time import time

from h5sh.utils import (make_column_kv_fmt, extract, format_shape)
from .base import Command
from .registry import register
###############################################################################


class Dump(Command):
    name = "dump"
    STDOUT = "STDOUT"

    def build_parser(self):
        parser = super(Dump, self).build_parser(
            description="Print the contents of a dataset.")
        parser.add_argument('-A', '--onlyattr',
                            help="Print only attributes",
                            action="store_true")
        parser.add_argument('-p', '--precision', type=int, default=4,
                            help="Floating point precision")
        parser.add_argument('-t', '--threshold', type=int, default=50,
                            help="Number of values to display")
        parser.add_argument('--suppress_small',
                            help="Print very small numbers as zero",
                            action="store_true")
        parser.add_argument('dataset', help="Dataset to print")
        parser.add_argument('-o', '--out',
                            help="File to save output",
                            default=Dump.STDOUT)
        return parser

    def execute(self, state, dataset, out, **kwargs):
        try:
            dataset = state.group[dataset]
        except KeyError:
            raise ValueError("Nonexistent dataset {!r}".format(dataset))

        if out == Dump.STDOUT:
            self._dump(dataset, sys.stdout, **kwargs)
        else:
            with open(out, 'w') as f:
                self._dump(dataset, f, **kwargs)

    def _dump(self, item, f, onlyattr, **kwargs):
        try:
            shape = item.shape
        except AttributeError:
            raise ValueError("{} is not a dataset".format(item.name))

        # Print size and shape
        f.write("Dataset: {}\n".format(item.name))
        f.write("Shape: {}".format(format_shape(shape)))
        if shape != item.maxshape:
            f.write("of max {}".format(format_shape(item.maxshape)))
        f.write("\n")

        # Print datatype
        dt = item.dtype
        if dt.names:
            # Write compound type in detail
            f.write("Type:\n")
            f.write(pformat(dt.descr))
        else:
            f.write("Type: {}\n".format(dt.name))

        # Print attributes
        attrs = dict(item.attrs)
        if attrs:
            # Convert byte-strings to strings
            attrs = dict((k, extract(v)) for (k, v) in attrs.items())
            f.write("Attributes:\n")
            f.write(pformat(attrs))
            f.write("\n")

        # Print chunking
        if item.chunks:
            f.write("Chunked: {}\n".format(format_shape(item.chunks)))

        if item.compression:
            f.write("Compressed: {}\n".format(item.compression))

        if onlyattr:
            return

        threshold = kwargs['threshold']
        if item.size > threshold:
            # More than 60 lines, roughly
            print("Truncating dataset ({:d} exceeds threshold {:d}): "
                  "use -t to increase".format(item.size, threshold), file=f)
            # Set edge items so that approximately "threshold" will appear:
            # e.g. if dimension is 3, then take cube root of half the threshold
            kwargs['edgeitems'] = max(1, int(pow(threshold/2, 1/len(shape))))

        print("---", file=f)

        item = extract(item)
        if shape:
            # Print array with given options
            print(np.array2string(item, **kwargs), file=f)
        else:
            print(item, file=f)


register.instance(Dump)

###############################################################################


class Attrs(Command):
    name = "attr"

    def build_parser(self):
        parser = super(Attrs, self).build_parser(
            description="Print attributes of the current group or a given "
            "object.")
        parser.add_argument('object', nargs='?')
        return parser

    def execute(self, state, obj=None):
        if obj is not None:
            obj = state.group[obj]
        else:
            obj = state.group

        attrs = obj.attrs
        keys = sorted(attrs)
        if not keys:
            return

        fmt = make_column_kv_fmt(keys, sep="=")
        for k in keys:
            v = to_native_str(attrs[k])
            print(fmt(k, v))


register.instance(Attrs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import h5py
import numpy as np

@pytest.fixture
def example_h5_filename(tmpdir, scope='module'):
    ext_filename = (tmpdir / "example-data-external.h5")
    filename = (tmpdir / "example-data.h5")

    with h5py.File(ext_filename,'w') as f:
        d = f.create_dataset("external_ds", data=np.array([1,2,3,4], dtype='d'))
        g = f.create_group("external_group")

    with h5py.File(filename,'w') as f:
        g = f.create_group("group")
        g.attrs['unit'] = "Flerbians"
        g.attrs['count'] = 123
        g2 = g.create_group("subgroup")
        g3 = g2.create_group("subsubgroup")

        d1 = g.create_dataset("scalar", data=1.23)
        d1.attrs['cats'] = np.array(["Kali", "Bustopher Jones"],
                                     dtype=h5py.special_dtype(vlen=str))

        d2 = g.create_dataset("vector", data=np.array([1,2,3], dtype='i'))
        f['link'] = d2
        f['softlink'] = h5py.SoftLink('/group/scalar')
        f['subsubgroup_hardlink'] = g3
        f['extlink'] = h5py.ExternalLink(ext_filename, "external_ds")
        f['extgroup'] = h5py.ExternalLink(ext_filename, "external_group")

    yield filename



# -*- coding: utf-8 -*-

"""Argument parser."""

from __future__ import (division, absolute_import, print_function, )
#-----------------------------------------------------------------------------#
from argparse import ArgumentParser
###############################################################################


class MiniSystemExit(BaseException):
    pass


class MiniArgParser(ArgumentParser):
    """Argument parser that just returns/raises instead of exiting.

    This utility class is used to parse arguments without actually exiting the
    python shell.
    """

    def __init__(self, *args, **kwargs):
        self.options = []
        self.dataset = False
        self.group = False
        super(MiniArgParser, self).__init__(*args, **kwargs)

    def exit(self, status=0, message=None):
        """Instead of calling sys.exit, raise an appropriate exception.
        """
        if status:
            raise TypeError(message)

        # If exit successful, just print the message
        raise MiniSystemExit(message)

    def error(self, message):
        """Instead of calling sys.exit, raise a type error.
        """
        raise TypeError(message)

    def add_dataset_argument(self, *args, **kwargs):
        self.dataset = True
        kwargs.setdefault('help', "Dataset name")
        return super(MiniArgParser, self).add_argument('dataset',
                                                       *args, **kwargs)

    def add_group_argument(self, *args, **kwargs):
        self.group = True
        kwargs.setdefault('help', "Group name")
        return super(MiniArgParser, self).add_argument('group',
                                                       *args, **kwargs)

    def add_object_argument(self, *args, **kwargs):
        self.dataset = True
        self.group = True
        kwargs.setdefault('help', "Object name")
        return super(MiniArgParser, self).add_argument('obj',
                                                       metavar='object', *args, **kwargs)

    def add_argument(self, name, *args, **kwargs):
        if name == 'dataset':
            return self.add_dataset_argument(*args, **kwargs)
        elif name == 'group':
            return self.add_group_argument(*args, **kwargs)
        elif name == 'object':
            return self.add_object_argument(*args, **kwargs)
        args = (name,) + args
        self.options.extend(opt for opt in args if opt.startswith('-'))
        return super(MiniArgParser, self).add_argument(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest

from h5sh.state import State
import h5sh.commands as module
from six import PY2

@pytest.fixture
def tmpstate(example_h5_filename):
    with State(example_h5_filename) as f:
        yield f

def test_interrupt(tmpstate, capsys):
    cmd = module.system.INTERRUPT_CMD
    tmpstate = None
    cmd(tmpstate)
    assert "\n" == capsys.readouterr().out
    cmd(tmpstate)
    assert "exit" in capsys.readouterr().out

def test_exit(tmpstate, capsys):
    cmd = module.COMMANDS['exit']
    tmpstate = None
    with pytest.raises(SystemExit):
        cmd(tmpstate)

def test_cd(tmpstate, capsys):
    cmd = module.COMMANDS['cd']

    # Into a group
    cmd(tmpstate, 'group')
    assert '/group' == tmpstate.cwd
    # Into a relative path
    cmd(tmpstate, '..')
    assert '/' == tmpstate.cwd
    cmd(tmpstate)
    assert '/' == tmpstate.cwd
    with pytest.raises(ValueError):
        cmd(tmpstate, 'link')
    with pytest.raises(ValueError):
        cmd(tmpstate, 'nonexistent')

def test_up(tmpstate, capsys):
    up = module.COMMANDS['up']
    cd = module.COMMANDS['cd']

    # Into a group
    cd(tmpstate, 'group')
    up(tmpstate)
    assert '/' == tmpstate.cwd

def test_ls(tmpstate, capsys):
    cmd = module.COMMANDS['ls']
    tmpstate.chdir('group')
    cmd(tmpstate)
    assert 'scalar subgroup vector\n' == capsys.readouterr().out

    cmd(tmpstate, '-l')
    assert """\
scalar   Dataset (d: scalar)
subgroup Group (1 item)
vector   Dataset (i: 3)
""" == capsys.readouterr().out

    cmd(tmpstate, '-1')
    assert 'scalar\nsubgroup\nvector\n' == capsys.readouterr().out

    tmpstate.chdir('/')
    cmd(tmpstate, '-l')
    assert """\
extgroup             Link (.../example-data-external.h5:external_group)
extlink              Link (.../example-data-external.h5:external_ds)
group                Group (3 items)
link                 Dataset (i: 3)
softlink             Link (/group/scalar)
subsubgroup_hardlink Group (0 items)
""" == capsys.readouterr().out

def test_dump(tmpstate, capsys, tmpdir):
    cmd = module.COMMANDS['dump']

    with pytest.raises(ValueError):
        cmd(tmpstate, 'group')

    cmd(tmpstate, '/group/scalar')
    out = capsys.readouterr().out
    if PY2:
        out = out.replace("u'", "'")
    assert """\
Dataset: /group/scalar
Shape: scalar
Type: float64
Attributes:
{'cats': array(['Kali', 'Bustopher Jones'], dtype=object)}
---
1.23
""" == out

    cmd(tmpstate, '/group/scalar', "-A")
    out = capsys.readouterr().out
    if PY2:
        out = out.replace("u'", "'")
    assert """\
Dataset: /group/scalar
Shape: scalar
Type: float64
Attributes:
{'cats': array(['Kali', 'Bustopher Jones'], dtype=object)}
""" == out

    cmd(tmpstate, '/group/vector')
    assert """\
Dataset: /group/vector
Shape: 3
Type: int32
---
[1 2 3]
""" == capsys.readouterr().out

    cmd(tmpstate, '-o', str(tmpdir / 'h5cmd_test_vector.txt'), '/group/vector')
    cmd(tmpstate, '-o', str(tmpdir / 'h5cmd_test_scalar.txt'), '/group/scalar')

def test_attr(tmpstate, capsys):
    cmd = module.COMMANDS['attr']

    cmd(tmpstate)
    assert "" == capsys.readouterr().out

def test_help(tmpstate, capsys):
    cmd = module.COMMANDS['help']
    tmpstate = None
    cmd(tmpstate)
    s = capsys.readouterr().out
    assert s.startswith("Available commands:")




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (division, absolute_import, print_function, )

from pprint import pprint
import pytest
import h5py
import sys

import h5sh.utils as module

@pytest.mark.skipif(sys.version_info < (3,3),
        reason="Test is for Python 3 only (unicode support")
def test_format_shape():
    format_shape = module.format_shape
    assert "scalar" == format_shape(())
    assert "2" == format_shape((2,))
    assert "2×10" == format_shape((2,10))
    assert "2×∞" == format_shape((2,None))

def test_shlex_split():
    split = module.shlex_split
    assert [] == split('')
    assert ['foo'] == split('foo')
    assert ['foo', 'bar'] == split('foo bar')
    assert ['foo bar'] == split(r'foo\ bar')
    assert ['foo bar', 'baz'] == split(r'"foo bar" baz')
    assert ['foo bar', 'baz'] == split(r'"foo bar" "baz')

def test_abspath():
    abspath = module.abspath

    assert "/foo/bar" == abspath("bar",   "/foo")
    assert "/" == abspath("/", "/foo")
    assert "/" == abspath("//////", "/foo")
    assert "/bar" == abspath("////bar", "/foo")
    assert "/bar" == abspath("bar",   "/")
    assert "/foo" == abspath("./foo", "/")
    assert "/foo" == abspath("./foo/", "/")
    assert "/foo" == abspath(".", "/foo")
    assert "/foo/bar" == abspath("..", "/foo/bar/baz")
    assert "/foo" == abspath("..", "/foo/bar")
    assert "/" == abspath("..", "/group")
    assert "/" == abspath("../..", "/foo/bar")
    assert "/foo" == abspath("../..", "/foo/bar/baz")
    assert "/foo" == abspath("/foo", "/bar/baz")

def test_subgroup(example_h5_filename):
    subgroup = module.subgroup
    with h5py.File(example_h5_filename, 'r') as example_h5_file:
        g = subgroup(example_h5_file, 'group')
        assert '/group' == g.name
        assert '/group' == subgroup(g, '.').name
        assert '/' == subgroup(g, '..').name
        assert '/group/subgroup' == subgroup(g, 'subgroup').name
        assert '/external_group' == subgroup(g, '/extgroup').name
        assert ('/subsubgroup_hardlink'
                == subgroup(g, '../subsubgroup_hardlink').name)
        with pytest.raises(ValueError):
            subgroup(example_h5_file, 'extlink')
        with pytest.raises(ValueError):
            subgroup(example_h5_file, 'nonexist')

def test_items(example_h5_filename):
    with h5py.File(example_h5_filename, 'r') as f:
        items = dict(module.items(f))
        assert sorted(items) == ['extgroup', 'extlink', 'group', 'link',
                'softlink', 'subsubgroup_hardlink']
        assert isinstance(items['group'], h5py.Group)
        assert isinstance(items['softlink'], h5py.SoftLink)
        assert isinstance(items['extgroup'], h5py.ExternalLink)
        assert isinstance(items['extlink'], h5py.ExternalLink)
        items = dict(module.items(f['group']))
        assert isinstance(items['scalar'], h5py.Dataset)

def test_short_describe(example_h5_filename):
    descr = module.short_describe
    with h5py.File(example_h5_filename, 'r') as f:
        items = dict((k,descr(v)) for (k,v) in module.items(f))
        # pprint(items)
        assert items == {
                'extgroup': 'Link (.../example-data-external.h5:external_group)',
                'extlink': 'Link (.../example-data-external.h5:external_ds)',
                'group': 'Group (3 items)',
                'link': 'Dataset (i: 3)',
                'softlink': 'Link (/group/scalar)',
                'subsubgroup_hardlink': 'Group (0 items)'}
        items = dict((k,descr(v)) for (k,v) in module.items(f['group']))
        assert items == {
                'scalar': 'Dataset (d: scalar)',
                'subgroup': 'Group (1 item)',
                'vector': 'Dataset (i: 3)'}

def test_unescape_string():
    un = module.unescape_string
    assert "This\nis Sparta" == un(r"This\nis Spart\x61")




# -*- coding: utf-8 -*-

from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from six import PY3
#-----------------------------------------------------------------------------#
import h5py
import numpy as np
import os
import shlex

###############################################################################
# STRING UTILITIES
###############################################################################


def shlex_split(s, comments=False, posix=True):
    """
    Splits a string using shell lexer, but returns any incomplete string as the
    last component instead of erroring for unmatched quotations.
    """
    lex = shlex.shlex(s, posix=posix)
    lex.whitespace_split = True
    if not comments:
        lex.commenters = ''

    result = []
    while True:
        try:
            tok = lex.get_token()
        except ValueError as e:
            print(repr(e))
            # Append the current token
            result.append(lex.token)
            break
        else:
            if tok == lex.eof:
                break
        result.append(tok)
    return result


def abspath(newpath, curpath):
    """Return the absolute path to the given 'newpath'.

    The current directory string must be given by 'curpath' as an absolute
    path.
    """
    assert newpath
    assert curpath
    assert curpath.startswith('/')

    subdirs = newpath.split('/')
    if not subdirs[0] or curpath == '/':
        # Absolute path (curpath is ignored)
        # or we're in the root directory
        dirs = [""]
    else:
        # Relative path; extract directory components from curpath
        dirs = curpath.split('/')

    for s in subdirs:
        if not s or s == ".":
            # Empty or 'current directory'
            pass
        elif s == "..":
            dirs.pop()
            if not dirs:
                raise ValueError("Too many '..' in path '{}'".format(newpath))
        else:
            dirs.append(s)

    if len(dirs) == 1:
        # Special case for root: joining [] or [""] return "", but you can't
        # set the first component to "/" since joining ["/","foo"] would
        # return "//foo"
        return '/'

    return '/'.join(dirs)


def subgroup(group, path):
    """Get a group inside another group, accounting for relative paths.

    This is the core functionality behind `chdir`.
    """
    assert path
    if '.' in path:
        path = abspath(path, group.name)

    try:
        group = group[path]
    except KeyError as e:
        raise ValueError("{}: No such group".format(path))
    except Exception as e:
        print("Failed to open group '{}' in '{}'".format(path, group.name))
        raise

    if not isinstance(group, h5py.Group):
        raise ValueError("{} is not a group (it is type {!s})".format(
            group.name, type(group).__name__))
    return group


def make_column_kv_fmt(keys, sep=" "):
    """Return the format function for a key/value pair that lines up all the
    keys and values.
    """
    maxlen = max(len(k) for k in keys) if keys else 1
    return "{{:{:d}s}}{:s}{{!s}}".format(maxlen, sep).format


if PY3:
    def unescape_string(text):
        return bytes(text, "utf-8").decode("unicode_escape")
else:
    def unescape_string(text):
        return text.decode("string_escape")

###############################################################################
# ARRAY UTILITIES
###############################################################################


def vectorized(dtype='O', nargin=1, nargout=1):
    """Decorator function for applying a python function to an array.

    Should have less overhead than ``np.vectorize``.
    """
    def _decorator(f):
        if np:
            _to_arr = np.frompyfunc(f, nargin, nargout)
        else:
            def _to_arr(x): return x

        if dtype == 'O':
            _vectorized = _to_arr
        else:
            def _vectorized(arr):
                result = _to_arr(arr)
                return np.array(result, dtype=dtype)
        return _vectorized
    return _decorator


if PY3:
    def to_native_str(s):
        if isinstance(s, bytes):
            # For ASCII-encoded attributes
            return s.decode('ascii')
        # For (possibly UTF-8 encoded attributes) or other types
        return s

    @vectorized(dtype='O')
    def to_native_str_array(s):
        return s.decode('ascii')

    @vectorized(dtype='S')
    def to_h5_str_array(s):
        return s.encode('ascii')
else:
    def to_native_str(s):
        #pylint: disable=undefined-variable
        if isinstance(s, unicode):
            return s.encode('ascii')
        # For native str datatypes or other types
        return s

    def to_native_str_array(arr):
        return arr

    def to_h5_str_array(arr):
        return arr

###############################################################################
# HDF5 UTILITIES
###############################################################################

_MULT_SYMBOL = u"\u00D7" if PY3 else "x"
_INF_SYMBOL = u"\u221E" if PY3 else "inf"


def format_shape(shape):
    """Return a nicely formatted description of a dataset shape.
    """
    if not shape:
        return "scalar"
    return _MULT_SYMBOL.join(str(i) if i is not None else _INF_SYMBOL
                             for i in shape)


def items(group):
    """Generator for key/value pairs in a group, returning links where
    possible.

    The only returned link types are External or Soft. Hard links are opened
    and the corresponding item is returned.
    """
    for key in group:
        value = group.get(key, getlink=True)
        if isinstance(value, h5py.HardLink):
            # Get the actual corresponding dataset or group
            value = group[key]
        yield (key, value)


def short_describe(obj):
    """Return a short description of the given group/dataset.
    """
    if isinstance(obj, h5py.Group):
        return "Group ({:d} item{:s})".format(len(obj),
                                              "s" if len(obj) != 1 else "")
    elif isinstance(obj, h5py.Dataset):
        return "Dataset ({:s}: {:s})".format(
            obj.dtype.char, format_shape(obj.shape))
    elif isinstance(obj, h5py.SoftLink):
        return "Link ({:s})".format(obj.path)
    elif isinstance(obj, h5py.ExternalLink):
        filename = obj.filename
        if os.path.isabs(filename):
            filename = os.path.sep.join(['...', os.path.basename(filename)])
        return "Link ({:s}:{:s})".format(filename, obj.path)
    elif isinstance(obj, h5py.HardLink):
        return "Object"
    else:
        # Unknown
        return str(obj.__class__)


def extract(data):
    """Extract data from h5py data objects.

    Note that scalars will be extracted as *numpy* scalars. To obtain a native
    Python scalar such as `float`, call `result.item()`.
    """
    try:
        shape = data.shape
    except AttributeError:
        # Not a numpy/h5py object
        return data

    if shape:
        # Extract array data (possibly compound)
        data = data[:]
        if PY3 and data.size and isinstance(next(data.flat), bytes):
            # In Python 3, variable-length ASCII strings are read as bytes,
            # which causes everything else in python 3 to be super unhappy.
            # Just convert it to strings.
            data = to_native_str_array(data)
    else:
        # Extract scalar data
        try:
            data = data[()]
        except (IndexError, AttributeError):
            # IndexError occurs when trying to get data from a np.void
            # dtype, i.e. a scalar compound datatype
            pass
        except TypeError:
            # In some versions, np.string_ won't accept the scalar slice
            data = to_native_str(data)
        else:
            data = to_native_str(data)

    return data



# -*- coding: utf-8 -*-

"""Registry of available commands."""

from __future__ import (division, absolute_import, print_function, )
#-----------------------------------------------------------------------------#
import sys
from functools import wraps

DEFINE_PARSER = ('argparse' in sys.modules)


class CommandRegistry(object):
    def __init__(self, commands=None):
        if commands is None:
            commands = {}
        self.commands = commands

    def __repr__(self):
        return "CommandRegistry({!r})".format(self.commands)

    def insert(self, key, func):
        if key in self.commands:
            raise KeyError("Duplicate command name '{}'".format(key))

        if DEFINE_PARSER:
            try:
                parser = func.parser
            except AttributeError:
                pass
            else:
                self._add_parser_cmd(key, parser)

        self.commands[key] = func

    def _add_parser_cmd(self, name, parser):
        """Define a free function that returns a function's parser.

        This is used by the 'sphinx-argparse' documentation generator.
        """
        module = sys.modules[__name__]
        attrname = "get_parser_" + name
        setattr(module, attrname, lambda: parser)

    def __iter__(self):
        return iter(self.commands)

    def __getitem__(self, key):
        return self.commands[key]


class RegisterCommand(object):
    def __init__(self, registry):
        self.registry = registry

    def __call__(self, description=None, name=None):
        """Add the following function using the function name as the command
        name.

        Example::

            @register(description="Print the current directory")
            def pwd(state):
                print(state.cwd)

        """
        def func(f, registry=self.registry, name=name, description=description):
            if name is None:
                name = f.__name__
            if description is not None:
                f.description = description
            registry.insert(name, f)
            return f

        return func

    def instance(self, cls, *args, **kwargs):
        """Add an *instance* of the given class to the command registry.

        Example::

            INTERRUPT_CMD = register.instance(Interrupt)

        """
        instance = cls(*args, **kwargs)
        self.registry.insert(instance.name, instance)
        return instance


# List of commands
COMMANDS = CommandRegistry()

# Register in the list of commands
register = RegisterCommand(COMMANDS)



# -*- coding: utf-8 -*-

"""Functions for navigating an HDF5 file."""


from __future__ import (division, absolute_import, print_function, )
#-----------------------------------------------------------------------------#
from h5sh.utils import (make_column_kv_fmt, short_describe, subgroup)

from .base import Command
from .registry import register

import h5py
###############################################################################


class Chdir(Command):
    name = "cd"

    def build_parser(self):
        parser = super(Chdir, self).build_parser(
            description="Change the current HDF5 group.")
        parser.add_argument('group', nargs='?')
        return parser

    def execute(self, state, group=None):
        try:
            state.chdir(group)
        except KeyError as e:
            raise ValueError(str(e))


cd = register.instance(Chdir)


@register("Print the path to the current HDF5 group")
def pwd(state):
    print(state.cwd)


class Up(object):
    def __init__(self, count):
        self.name = "u" + "p" * count
        self.path = "../" * count
        self.description = "Traverse up {:d} directories.".format(count)

    def __call__(self, state):
        cd(state, self.path)


for _i in range(1, 6):
    register.instance(Up, _i)
del _i


class Listdir(Command):
    name = "ls"

    def build_parser(self):
        parser = super(Listdir, self).build_parser(
            description="List items in the current group.")
        parser.add_argument('-l', dest='long', action='store_true',
                            help="Print attributes as well as names")
        parser.add_argument('-1', dest='oneline', action='store_true',
                            help="Print one entry per line")
        parser.add_argument('group', nargs='?')
        return parser

    def execute(self, state, long, oneline, group=None):
        if group is not None:
            group = subgroup(state.group, group)
        else:
            group = state.group

        keys = sorted(group)
        if not keys:
            return

        if long:
            fmt = make_column_kv_fmt(keys)
            for k in keys:
                v = group.get(k, getlink=True)
                if isinstance(v, h5py.HardLink):
                    v = group[k]
                v = short_describe(v)
                print(fmt(k, v))
        elif oneline:
            for k in keys:
                print(k)
        else:
            print(" ".join(keys))


ls = register.instance(Listdir)


@register("Alias for 'ls -l'")
def l(state, *args):
    ls(state, "-l", *args)



# -*- coding: utf-8 -*-

from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
#-----------------------------------------------------------------------------#
import os

FILENAME = 'class:filename'
TRUNC_FILENAME = 'class:filename.trunc'
HDF5_GROUP = 'class:hdf5.group'
PROMPT_TOKEN = 'class:prompt'

STYLES = [
    (FILENAME, 'fg:cyan'),
    (TRUNC_FILENAME, 'fg:gray'),
    (HDF5_GROUP, 'fg:green'),
    (PROMPT_TOKEN, 'bold'),
]


def get_style_rules():
    return [(k.replace('class:', ''), v) for (k, v) in STYLES]


def styled_filename(filename, width=32):
    """Filename, truncated to the given width, as a list to be wrapped by
    prompt_toolkit.FormattedText.
    """
    def len_le_width(result):
        return sum(len(v[1]) for v in result) <= width

    result = [(FILENAME, filename)]
    if len_le_width(result):
        # If the full path is small, just return it
        return result

    # See if we can strip the leading directories
    basename = os.path.basename(filename)
    if basename != filename:
        # Shortened the directory
        dirname = [(TRUNC_FILENAME, ".../")]
        result = dirname + [(FILENAME, basename)]
    else:
        dirname = []

    if len_le_width(result):
        # If the full path is small, just return it
        return result

    # Length for half of basename, subtracting len('.../') + len('...')
    # with minimum in case width < 16
    half_width = max(4, (width - 7) // 2)
    result = dirname + [
        (FILENAME, basename[:half_width]),
        (TRUNC_FILENAME, "..."),
        (FILENAME, basename[-half_width:])]
    return result



# -*- coding: utf-8 -*-

from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
#-----------------------------------------------------------------------------#
from contextlib import contextmanager
import h5py
import os
import sys

from .utils import abspath
from .styles import (styled_filename, HDF5_GROUP, PROMPT_TOKEN)

###############################################################################


class State(object):
    """The state of the current "shell".

    At the moment this merely encapsulates the working directory and the open
    file.
    """

    def __init__(self, filename, mode='r'):
        # HDF5 file
        self.f = h5py.File(filename, mode)
        # Current group
        self.group = self.f
        # Groups/datasets inside the current group
        self._cur_items = None

    @property
    def subgroups(self):
        """Get a cached list of groups inside the current group.
        """
        if self._cur_items is None:
            self._update_cur_items()
        return self._cur_items[0]

    @property
    def datasets(self):
        """Get a cached list of datasets inside the current group.
        """
        if self._cur_items is None:
            self._update_cur_items()
        return self._cur_items[1]

    def _update_cur_items(self):
        _cur_group = self.group
        groups = []
        datasets = []
        for key in _cur_group:
            cls = _cur_group.get(key, getclass=True)
            if issubclass(cls, h5py.Group):
                groups.append(key)
            elif issubclass(cls, h5py.Dataset):
                datasets.append(key)
        self._cur_items = (groups, datasets)

    def close(self):
        self.f.close()
        self.f = None

    @property
    def closed(self):
        return self.f is None

    def chdir(self, dir=None):
        if dir is None:
            # Return to base directory
            self.group = self.f
            return

        if '.' in dir:
            dir = abspath(dir, self.cwd)

        group = self.group[dir]
        if not isinstance(group, h5py.Group):
            raise ValueError("{} is not a group".format(group.name))
        self.group = group
        # Clear cache of current items
        self._cur_items = None

    @property
    def filename(self):
        """Filename of the file"""
        return self.f.filename

    @property
    def cwd(self):
        """Path of the current HDF5 group"""
        return self.group.name

    def get_styled_prompt(self, output):
        """Get a list [(clsfmt, text), ...] for the prompt.
        """
        cols = output.get_size().columns
        cwd = self.cwd
        # Try to fit the prompt on no more than half the terminal width
        max_filename_len = cols // 2 - (len(cwd) + 4)

        result = styled_filename(self.filename, max_filename_len) + [
            ('', ":"),
            (HDF5_GROUP, self.cwd),
            (PROMPT_TOKEN, ' > '),
        ]

        return result

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

###############################################################################
# end of Nemesis/python/exnihilotools/h5sh/state.py
###############################################################################



# -*- coding: utf-8 -*-

"""Registry of available commands."""

from __future__ import (division, absolute_import, print_function, )
#-----------------------------------------------------------------------------#
from .miniargparse import MiniArgParser, MiniSystemExit


class Command(object):
    """Utility class for constructing a command that takes arguments.
    """
    name = None

    def __init__(self):
        self.parser = self.build_parser()

    def build_parser(self, **kwargs):
        parser = MiniArgParser(prog=self.name, **kwargs)
        return parser

    def execute(self, state, *args, **kwargs):
        raise NotImplementedError()

    def __call__(self, state, *args):
        # Parse the arguments
        try:
            parsed = self.parser.parse_args(args)
        except MiniSystemExit:
            return

        self.execute(state, **vars(parsed))

    def get_completions(self, document, args, state):
        """Get completions for this command.

        'args' is the list of given arguments to the command.
        """
        in_word = document.get_word_before_cursor(WORD=True)
        if in_word.startswith('-'):
            for arg in self.parser.options:
                yield arg
        else:
            if self.parser.dataset:
                for arg in state.datasets:
                    yield arg
            if self.parser.group:
                for arg in state.subgroups:
                    yield arg

    @property
    def description(self):
        desc = self.parser.description
        if not desc:
            raise AttributeError(desc)
        return desc



# -*- coding: utf-8 -*-

"""Command-line utilities subsystem."""

# Command utilities
from .base import Command
from .miniargparse import MiniArgParser
from .registry import COMMANDS, register

# Load actual commands
from . import navigation
from . import query
from . import system



# -*- coding: utf-8 -*-

"""h5sh system utility functions."""

from __future__ import (division, absolute_import, print_function, )
#-----------------------------------------------------------------------------#
import sys
from time import time

from h5sh.utils import (make_column_kv_fmt, short_describe, unescape_string)

from .base import Command
from .registry import register, COMMANDS
###############################################################################


class Interrupt(object):
    name = "__INTERRUPT__"
    warning_time = 2.0

    def __init__(self):
        self.last_time = 0

    def __call__(self, state):
        cur_time = time()
        if cur_time - self.last_time < Interrupt.warning_time:
            # Print helpful message on two consecutive interrupts within 2
            # seconds
            print("\n(Use the 'exit' command or type Ctrl-D to quit)")
        else:
            print()
        self.last_time = cur_time


INTERRUPT_CMD = register.instance(Interrupt)

###############################################################################


@register(name="__NULL__")
def null(state):
    pass


NULL_CMD = null

###############################################################################


@register(name="exit", description="Exit h5sh.")
def _exit(state):
    sys.exit(0)

###############################################################################


@register(name='help',
          description="List available commands.")
def _help(state):
    print("Available commands:")
    keys = sorted(k for k in COMMANDS if not k.startswith("__"))
    fmt = make_column_kv_fmt(keys, " - ")
    for k in keys:
        cmd = COMMANDS[k]
        try:
            desc = cmd.description
        except AttributeError:
            desc = ""

        print(fmt(k, desc))

###############################################################################


@register("Print the name of the file being examined")
def filename(state):
    print(state.f.filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Set up and install h5sh."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

setup(
    author="Seth Robert Johnson",
    author_email="johnsonsr@ornl.gov",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Fast shell-like tool for interacting with HDF5 files.",
    entry_points={"console_scripts": ["h5sh=h5sh.scripts.main:main"],},
    include_package_data=True,
    install_requires=[
        "h5py>=2.7.1",
        "prompt-toolkit>=2.0",
        "numpy>=1.15",
        "pygments",
        "six",
    ],
    license="BSD license",
    long_description=readme + "\n\n" + history,
    keywords="h5sh",
    name="h5sh",
    packages=find_packages(include=["h5sh", "h5sh.*"]),
    test_suite="tests",
    tests_require=["pytest"],
    url="https://github.com/sethrj/h5sh",
    version="0.1.1",
    zip_safe=False,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (division, absolute_import, print_function,
        unicode_literals)

from pprint import pprint
import pytest
import h5py

import h5sh.styles as module

def test_styled_filename():
    style = module.styled_filename
    FN = module.FILENAME
    TFN = module.TRUNC_FILENAME

    assert style("foo.h5") == [(FN, 'foo.h5')]
    assert style("abcd"*8 + ".h5") == [
            (FN, 'abcdabcdabcd'), (TFN, '...'), (FN, 'dabcdabcd.h5')]
    assert style("x"*32 + "/foo.h5") == [
            (TFN, '.../'), (FN, 'foo.h5')]
    assert style("x"*32 + "/" + "abcd"*8 + ".h5") == [
            (TFN, '.../'), (FN, 'abcdabcdabcd'), (TFN, '...'),
            (FN, 'dabcdabcd.h5')]
    assert style("basename/" + "abcd"*8 + ".h5") == [
            (TFN, '.../'), (FN, 'abcdabcdabcd'), (TFN, '...'),
            (FN, 'dabcdabcd.h5')]




# -*- coding: utf-8 -*-

from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from six.moves import input
#-----------------------------------------------------------------------------#
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.styles import Style

from .commands import COMMANDS
from .commands.system import INTERRUPT_CMD, NULL_CMD
from .styles import get_style_rules
from .utils import shlex_split

###############################################################################


def _get_console_lexer():
    try:
        from prompt_toolkit.lexers import PygmentsLexer
        from pygments.lexers import BashLexer
    except ImportError:
        return None
    return PygmentsLexer(BashLexer)


class CommandCompleter(Completer):
    def __init__(self, state):
        self.state = state
        self.cmd_names = sorted(k for k in COMMANDS if not k.startswith('_'))

    def get_completions(self, document, complete_event):
        preceding_word = document.get_word_before_cursor(WORD=True)

        if len(document.text) == len(preceding_word):
            # Empty or completing the first command: list commands that start
            # with the prefix
            completions = self.cmd_names
        else:
            # Process the text that's there
            args = shlex_split(document.text)
            try:
                cmd = COMMANDS[args[0]]
            except KeyError:
                return

            try:
                get_completions = cmd.get_completions
            except AttributeError:
                return

            completions = cmd.get_completions(document, args[1:], self.state)

        pos = -len(preceding_word)
        for arg in completions:
            if arg.startswith(preceding_word):
                yield Completion(arg, pos)


class Console(object):
    def __init__(self, state):
        # Command-line state
        self.state = state
        # Debug mode
        self.debug = False
        # Prompt session
        self.session = PromptSession(lexer=_get_console_lexer(),
                                     style=Style(get_style_rules()),
                                     completer=CommandCompleter(state))
        # prompt_toolkit Output class
        self.output = self.session.app.output

    def prompt(self):
        with patch_stdout(raw=True):
            prompt = FormattedText(self.state.get_styled_prompt(self.output))
            text = self.session.prompt(prompt)
        return text

    def read(self):
        """Prompt for and read a single command.
        """
        try:
            text = self.prompt()
        except KeyboardInterrupt:
            cmd = INTERRUPT_CMD.name
            args = ()
        except EOFError:
            cmd = "exit"
            args = ()
        else:
            args = shlex_split(text)
            if args:
                cmd = args[0]
                args = args[1:]
            else:
                (cmd, args) = ("__NULL__", ())
        return (cmd, args)

    def interact(self):
        while True:
            (cmd_name, args) = self.read()
            try:
                cmd = COMMANDS[cmd_name]
            except KeyError:
                print("h5sh: {}: command not found".format(cmd_name))
                continue

            try:
                cmd(self.state, *args)
            except KeyboardInterrupt:
                INTERRUPT_CMD(self.state)
            except Exception as e:
                if not self.debug and isinstance(e, (TypeError, ValueError)):
                    print("{}: {!s}".format(cmd_name, e))
                    continue

                import logging as log
                log.exception(e)
                continue
